/*++

(C) Copyright 2000-2007 Hewlett-Packard Development Company, L.P.

Module Name:

    CpqCiKlx.c

Abstract:

    This module contains the code for a ring-0 Linux driver that
    provides kernel mode services to an application in user space.

    This code is the "wapper" for the common library used to access
    The HP Integrate Lights-Out (iLO) controller.

Environment:

    Kernel mode

Revision History:

--*/


#include "CpqCiHlx.h"

#ifdef CPQCI_X86_64 
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
#include <linux/compat.h>
#else
#include <asm/ioctl32.h>
#endif
#endif

#include "CpqCiWrp.h"
#include "CpqCiTyp.h"
#include "CpqCiSem.h"
#include "CpqCi.h"
#include "CpqCiKlx.h"
#include "CpqCiUlx.h"

#define HPQCI_DESCRIPTION "This module support the HP iLO & iLO 2 management controllers"

MODULE_DESCRIPTION(HPQCI_DESCRIPTION);
MODULE_AUTHOR("Hewlett-Packard Development Company, L.P.");
MODULE_LICENSE("GPL");

#define DEVICE_NAME "hpqci"

static int hpqci_pci_probe(struct pci_dev *pdev, const struct pci_device_id *ent);
static void hpqci_pci_shutdown(struct pci_dev *pdev);

#ifdef NOT_NEEDED
static struct device_driver hpqci_base_driver =
{
    .name = DEVICE_NAME,
    .bus = &platform_bus_type
};
#endif



static const struct pci_device_id hpqci_dev[] = { 
    { PCI_DEVICE(CPQCI_VEND_ID, CPQCI_DEV_ID), },
    { 0 }
};

MODULE_DEVICE_TABLE(pci, hpqci_dev);

static struct pci_driver hpqci_pci_driver = {
    .name       =   DEVICE_NAME,
    .id_table   =   hpqci_dev,
    .probe      =   hpqci_pci_probe,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
    .shutdown   =   hpqci_pci_shutdown
#endif
};


/*************************************************************************
 ** Cross Archhitecture and Linux Release Compatibility Section
 *************************************************************************/

#ifdef CONFIG_X86_64
#if LINUX_VERSION_CODE >=KERNEL_VERSION(2,6,10)
static long cpqci_compat_ioctl(struct file *f, unsigned int cmd, unsigned long arg);
int cpqci_ioctl32_handler(struct file *file, unsigned int cmd, unsigned long arg);
#else
extern asmlinkage long sys_ioctl(unsigned int fd, unsigned int cmd, void *);
int cpqci_ioctl32_handler(unsigned int fd, unsigned int cmd, unsigned long arg, struct file *file); 
#endif
#endif  /* CONFIG_X86_64 */

#ifndef IRQF_DISABLED
#define IRQF_DISABLED	SA_INTERRUPT
#endif

#ifndef IRQF_SHARED
#define IRQF_SHARED	SA_SHIRQ
#endif


#ifdef  DEBUG 
void DbgDump(struct _CPQCIADAPTEROBJECT* Adapter, CPQCICONNECTION Connection);
#else
#define DbgDump( Adapter, Connection )
#endif

#define CPQCI_CONNECTION_NAME  "HpqCiCO"

#define MS_TO_JIFFIES(ms)   ((ms)/(1000/HZ))

//
//  Driver support prototypes for this module.
//
static int cpqci_open(struct inode *i, struct file *f);
static int cpqci_flush(struct file *f, fl_owner_t id);
static unsigned int cpqci_poll(struct file *f, poll_table *wait);
static int cpqci_fasync(int fd, struct file *f, int on);
static int cpqci_close(struct inode *i, struct file *f);
static int cpqci_ioctl(struct inode *i, struct file *f, unsigned int cmd,
               unsigned long arg);
static int cpqci_mmap(struct file *f, struct vm_area_struct *vma);
irqreturn_t cpqci_intr(int irq, void *dev_id, struct pt_regs *regs);
void CpqCiDrvOsFree(CPQCICONTEXT Context, CPQCIDRVOSHANDLE MemoryHandle);



CPQCIDRVOSHANDLE CpqCiDrvOsMallocContiguous(CPQCICONTEXT Context, CPQCISIZE AllocationSize);
CPQCIDRVOSHANDLE CpqCiDrvOsMap(CPQCICONTEXT Context, CPQCIPHYSICAL PhysicalAddress, CPQCISIZE MappedSize, 
        CPQCIMODE AddressMode, CPQCICACHETYPE CacheType);
CPQCISTATUS CpqCiOsGetRefCount( CPQCICONTEXT Context, CPQCITYPEQUAD* RefCount);

void CpqCiLockAdapter(void);
void CpqCiUnlockAdapter(void);



//
// Externs 
//
extern struct proc_dir_entry* cpqci_entry_print_cpqci;
extern int cpqci_read_print_cpqci(char *Buf, char **start, off_t off, int count, int *eof, void *data);
extern int CpqCiShareContext( CPQCICONTEXT Context, struct _CPQCIADAPTEROBJECT *pao,  CPQCICONNECTION connection);
extern CPQCIDRVOSHANDLE GetMemoryHandle( struct _CPQCIADAPTEROBJECT *pao,  CPQCICONNECTION connection);

//
// Globals
//
CPQCIADAPTEREXTENSION *CpqCi;

int  LockAdapterCount;


static struct file_operations cpqci_ops = {
    THIS_MODULE,
    open:       cpqci_open,
//    flush:      cpqci_flush,
    release:    cpqci_close,
    ioctl:      cpqci_ioctl,
//    fasync:     cpqci_fasync,
//    poll:       cpqci_poll,
    mmap:       cpqci_mmap,
#if defined(CPQCI_X86_64) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
    compat_ioctl:   cpqci_compat_ioctl,
#endif

};

static struct miscdevice cpqci_miscdev = {
    CPQCI_MINOR,
    "cpqci",
    &cpqci_ops,
};

// VMA operations for shared memory
static void cpqci_vma_open(struct vm_area_struct *vma);
static void cpqci_vma_close(struct vm_area_struct *vma);

static void cpqci_do_munmap(CPQCICONTEXT Context, CPQCIPHYSICAL VirtualAddress, CPQCISIZE MappedSize);

static struct vm_operations_struct cpqci_vm_ops_regular = {
    open:   cpqci_vma_open,
    close:  cpqci_vma_close,
    nopage: cpqci_vma_nopage_regular,
};

static struct vm_operations_struct cpqci_vm_ops_high = {
    open:  cpqci_vma_open,
    close: cpqci_vma_close,
    nopage: cpqci_vma_nopage_high,
};




//========================================================
// Managing the Adapter Lock to synchronize Adapter access
//========================================================


void CpqCiLockAdapter(void)    // Uses Global ADAPTEREXTENSION *CpqCi
{
   if (LockAdapterCount > 0)
      printk(KERN_DEBUG "LOCKING ADAPTER:  LOCK ALREADY TAKEN!!!\n");
   spin_lock_irqsave(&CpqCi->AdapterLock, CpqCi->AdapterLockFlags);
   LockAdapterCount++;
}


void CpqCiUnlockAdapter(void)    // Uses Global ADAPTEREXTENSION *CpqCi
{
    if (LockAdapterCount <= 0)
      printk(KERN_ERR "UNLOCKING ADAPTER - WAS NOT LOCKED\n\n");
   LockAdapterCount--;
   spin_unlock_irqrestore(&CpqCi->AdapterLock, CpqCi->AdapterLockFlags);
}


//========================================================
// Managing the Adapter Lock to synchronize Channel access
//========================================================


void CpqCiLockChannel(int channel)    // Uses Global ADAPTEREXTENSION *CpqCi
{
//   DBG("LOCKING CHANNEL:  %d\n", channel);
   spin_lock_irqsave(&CpqCi->ChannelLock[channel], CpqCi->ChannelLockFlags[channel]);
}


void CpqCiUnlockChannel(int channel)  // Uses Global ADAPTEREXTENSION *CpqCi
{
//   DBG("UNLOCKING CHANNEL:  %d\n\n", channel);
   spin_unlock_irqrestore(&CpqCi->ChannelLock[channel], CpqCi->ChannelLockFlags[channel]);
}




//====================================================
//Creating, Deleting and Finding Connection Extensions
//====================================================

//A semaphore to guide the list of connections and
//also make sure close_helper is called in a serial fashion.
struct semaphore connection_lock; 

//We can assume that the connection lock is engaged
static CPQCICONNECTIONEXTENSION* cpqci_new_connection(void) {

    CPQCICONNECTIONEXTENSION* c;

    c = kmalloc(sizeof(CPQCICONNECTIONEXTENSION), GFP_KERNEL);
    DBG("(CONNECTIONEXTENSION)kmalloc(%d) = %p\n", sizeof(CPQCICONNECTIONEXTENSION), c);
    if (!c) return NULL;
    
    memset(c, 0, sizeof(CPQCICONNECTIONEXTENSION));
    if (!CpqCi) return c;

    c->Next = CpqCi->ConnectionList;
    CpqCi->ConnectionList = c;
    DBG("Insert CONNECTIONOBJECT into chain and returning %p\n", c);

    return c;
}


//We can assume that the connection lock is engaged
static CPQCICONNECTIONEXTENSION** cpqci_find_connection(CPQCICONNECTIONEXTENSION* c) {

    CPQCICONNECTIONEXTENSION** pc;

    if (!CpqCi) return NULL;

    pc = &CpqCi->ConnectionList;
    while (*pc) {
        DBG("Matching Connection Object %p against %p\n", (*pc), c);
        if ((*pc) == c) break; 
        pc = &((*pc)->Next);
    }

    return pc;
}

//We can assume that the connection lock is engaged
static void cpqci_delete_connection_mem_handle(CPQCICONNECTIONEXTENSION* c) {

        if (!c) return; 
    //If the buffer was used, we gotta clean up
    if (c->MemHandleMapped) {
        DBG("(connection) free MemHandleMapped %p (virt %p, phy %p)\n", 
                c->MemHandleMapped,
                c->MemHandleMapped->VirtualAddress,
                c->MemHandleMapped->PhysicalAddress);
        //Context is not needed !! (the memhandle has it all)
        CpqCiDrvOsFree(NULL, c->MemHandleMapped);
        c->MemHandleMapped = NULL;
    }
    if (c->MemHandleMalloc) {
        DBG("(connection) free MemHandleMalloc %p (virt %p, phy %p)\n", 
                c->MemHandleMalloc,
                c->MemHandleMalloc->VirtualAddress,
                c->MemHandleMalloc->PhysicalAddress);
        //Context is not needed !! (the memhandle has it all)
        CpqCiDrvOsFree(NULL, c->MemHandleMalloc);
        c->MemHandleMalloc= NULL;
    }
}

//We can assume that the connection lock is engaged
static void cpqci_new_connection_mem_handle(CPQCICONTEXT Context, CPQCICONNECTIONEXTENSION* c, CPQCISIZE AllocationSize) {

        if (!c) return; 
    //If the buffer was used, we are not going to proceed 
    if (c->MemHandleMapped || c->MemHandleMalloc) return; 
    
    //Allocate kernel memory and map it
    c->MemHandleMalloc = CpqCiDrvOsMallocContiguous(Context, AllocationSize);
    if (c->MemHandleMalloc) {
#ifdef DEBUG
        memset(c->MemHandleMalloc->VirtualAddress, 0xA5, AllocationSize);
#else
        memset(c->MemHandleMalloc->VirtualAddress, 0x0, AllocationSize);
#endif

        DBG("(connection) alloc MemHandleMalloc %p (virt %p, phy %p)\n", 
                c->MemHandleMalloc,
                c->MemHandleMalloc->VirtualAddress,
                c->MemHandleMalloc->PhysicalAddress);
        c->MemHandleMapped = CpqCiDrvOsMap(Context, c->MemHandleMalloc->PhysicalAddress, 
                AllocationSize, CpqCiUserMode, CpqCiCached );
    }
    if (c->MemHandleMapped == NULL) {
        CpqCiDrvOsFree(NULL, c->MemHandleMalloc);
        c->MemHandleMalloc = NULL;
    } else {
        DBG("(connection) alloc MemHandleMapped %p (virt %p, phy %p)\n", 
                c->MemHandleMapped,
                c->MemHandleMapped->VirtualAddress,
                c->MemHandleMapped->PhysicalAddress);
    }
}

//We can assume that the connection lock is engaged
static void cpqci_delete_connection(CPQCICONNECTIONEXTENSION* c) {

    CPQCICONNECTIONEXTENSION** pc;
      
        if (!c) return; 
    pc = cpqci_find_connection(c);
    if (!pc) return;

    if (*pc != NULL) {
        DBG("Delete CONNECTIONOBJECT from chain\n");
        *pc = c->Next;
    }

    cpqci_delete_connection_mem_handle(c);
    DBG("(cpqci_delete_connection)kifree(%p)\n", c);
    kfree(c);
}


//====================================================
//Cleanup routines 
//====================================================
int cpqci_close_helper(CPQCICONTEXT Context, CPQCICONNECTIONEXTENSION *connectionExtension)
/*++

Routine Description:

    This routine is called as a result of a user call to CloseHandle() or
    program termination.
    
Arguments:

    connectionExtension -

Return Value:

    The result of the operation.

--*/
{
    CPQCIADAPTEREXTENSION *adapterExtension = CpqCi;
    int connection;
    CPQCICHANNEL channel;
    CPQCISTATUS error;
        unsigned long long ref;
    struct file* f = (struct file*) Context;

    DBG("cpqci_close_helper ENTER\n");
    CpqCiLockAdapter();
    
    if (connectionExtension == NULL) {
    	DBG("cpqci_close_helper:  connectionExtension == NULL!! Early EXIT\n");
        CpqCiUnlockAdapter();
        return 0;
    }

    CpqCiOsGetRefCount(Context, (CPQCITYPEQUAD*) &ref);
    DBG(">>>pid %d: Connection %d ref %ld\n", current->pid, connectionExtension->Connection, ref);

    // We need no longer need to hold the lock.
    CpqCiUnlockAdapter();
    
    //  We only want to close a connection.
    if (connectionExtension->Header.ObjectType == CpqCiConnection) {

        connection = connectionExtension->Connection;
        //  Close the channel.
        error = CpqCiDrvCloseChannel(Context, adapterExtension->Adapter,
                connection, &channel);


        //error should be ChannelStillInuse or ConnectionStillInuse or OK.
        if (error != CpqCiErrConnectionStillInuse) {
            if (error != CpqCiErrChannelStillInuse) {
                //  Signal the next waiter on the specified priority queue
                //  that a channel has become available.
                if (channel < CPQCI_BASE_HIGHPRIORITY_CHANNEL)
                    up(&adapterExtension->ChannelSem[0]);
                else if (channel < CPQCI_BASE_CRITICALPRIORITY_CHANNEL)
                        up(&adapterExtension->ChannelSem[1]);

                DBG("pid: %d *** Free Channel %d\n", current->pid, channel);
            }
        
            //This will close the connection extension object of the OS 
            DBG("File %p private data gets deleted\n", f);
            if (f) f->private_data = NULL;  
            DBG("connection extension %p gets deleted\n", connectionExtension);
            cpqci_delete_connection(connectionExtension);

            DBG("pid: %d *** Free Connection %d\n", current->pid, connection);
        } else {
            DBG ("pid: %d >>> Connection %d & Channel %d still in use.\n", current->pid, connection, channel);
        }

        if (f == NULL) {
            DBG (">>> connection extension %p gets deleted\n", connectionExtension);
            cpqci_delete_connection(connectionExtension);
        }

    }

#ifndef DEBUG
    //MOD_DEC_USE_COUNT;
#endif
    DBG(">>pid %d: close_helper finished\n", current->pid);
    DBG("cpqci_close_helper:  NORMAL EXIT\n");
    return 0;
}

void cleanup_module_helper(int irq_status, int irq, int register_status, struct _CPQCIADAPTEROBJECT* adapter, CPQCIADAPTEREXTENSION *adapterExtension)
/*
 * This is to clean up a (partially) registered channel driver
 *
 *
 */
{
    CPQCICONNECTIONEXTENSION* ci_conn;

    DBG("cleanup_module_helper ENTER\n");

#if defined(CPQCI_X86_64) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
    unregister_ioctl32_conversion(CPQCI_IOCTL_PHYSMEM_ALLOC);
    unregister_ioctl32_conversion(CPQCI_IOCTL_PHYSMEM_FREE);
    unregister_ioctl32_conversion(CPQCI_IOCTL_CREATE_CHANNEL);
    unregister_ioctl32_conversion(CPQCI_IOCTL_OPEN_CHANNEL);
    unregister_ioctl32_conversion(CPQCI_IOCTL_REGISTER_WAITER);
    unregister_ioctl32_conversion(CPQCI_IOCTL_GET_INFO);
    unregister_ioctl32_conversion(SEM_CREATE);
    unregister_ioctl32_conversion(SEM_DESTROY);
    unregister_ioctl32_conversion(SEM_UP);
    unregister_ioctl32_conversion(SEM_DOWN);
    unregister_ioctl32_conversion(SEM_DOWN_TIMER);
    unregister_ioctl32_conversion(SEM_GET_VALUE);
    unregister_ioctl32_conversion(SEM_SET_VALUE);
#endif

#if defined(CONFIG_PROC_FS)
    if (cpqci_entry_print_cpqci){
            remove_proc_entry("hpqci", &proc_root);
    }
#endif 

    //This is a chunk of memory for semaphore handling
    DBG(" Calling global_sem_destroy\n");
    cpqci_global_sem_destroy();

    //  We should remove the interrupt service routine first.  This keeps
    //  the ISR off of the event object.
    DBG("Calling free_irq (if %d > 0): adapterExtension  %p\n", irq, adapterExtension);
    if (irq_status == 0 && irq > 0)
        free_irq(irq, adapterExtension);

    if (register_status == 0)
        misc_deregister(&cpqci_miscdev);

    // Kill any open connection left
    ci_conn = (adapterExtension == NULL) ? NULL : adapterExtension->ConnectionList; 
    DBG("Delete any remaining open connections....\n");
    while (ci_conn != NULL) {
        CPQCICONNECTIONEXTENSION* prev_ci_conn = ci_conn;
        DBG("Deleting connection extension %p\n", prev_ci_conn);
        ci_conn = ci_conn->Next;
        //This "unshares" the connection object (context is NULLED).
        DBG("Calling CpqCiShareContext\n");
        CpqCiShareContext(NULL, adapter, prev_ci_conn->Connection);
	DBG("Calling cpqci_close_helper\n");
        cpqci_close_helper(NULL, prev_ci_conn);
    }
    DBG("Done");

    //  Next we delete any adapter resources created.
    if (adapter)
        CpqCiDrvDeleteAdapter(NULL, adapter);

    //  If the adapter device object was created, then we have other 
    //  resources to wack.
    if (adapterExtension) {
        DBG("(cleanup_module_helper) adapterExtension (CpqCi) kfree(%p)\n", adapterExtension);
        kfree(adapterExtension);
    }

    DBG("cleanup_module_helper EXIT\n");
}


int cpqci_open(struct inode *i, struct file *f)
/*++

Routine Description:

    This routine is called as a result of a user call to CreateFile().
    
Arguments:

    i - inode of open device node
    f - file handle of open device node

Return Value:

    The result of the operation.

--*/
{
#ifndef DEBUG
    //MOD_INC_USE_COUNT;
#endif
    DBG("cpqci_open pid = %d\n", current->pid);
    DBG("cpqci_open file= %p\n", f);
    DBG("cpqci_open file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);
    f->private_data = NULL;
    return 0;
}



int cpqci_flush(struct file *f, fl_owner_t id)
/*++

Routine Description:

    This routine is called as a result of a user call to CloseHandle() or
    program termination.
    
Arguments:

    i - inode of open device node
    f - file handle of open device node

Return Value:

    The result of the operation.

--*/
{
    CPQCICONNECTIONEXTENSION *connectionExtension;

    if (!f) return 0;
    DBG("cpqci_flush file= %p\n", f);
    DBG("cpqci_flush pid = %d\n", current->pid);
    DBG("cpqci_flush (private = %p)\n", f->private_data);
    DBG("cpqci_flush file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);

    connectionExtension = f->private_data;
    //f->private_data = NULL;
    if (connectionExtension == NULL)
        return 0;

    return cpqci_close_helper(f, connectionExtension);
    //return cpqci_close_helper2(f);

}

int cpqci_close(struct inode *i, struct file *f) {
    CPQCICONNECTIONEXTENSION *connectionExtension;

    if (!f) return 0;
    DBG("cpqci_close file= %p\n", f);
    DBG("cpqci_close pid = %d\n", current->pid);
    DBG("cpqci_close (private = %p)\n", f->private_data);
    DBG("cpqci_close file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);

    connectionExtension = f->private_data;
    //f->private_data = NULL;
    if (connectionExtension == NULL)
        return 0;

/*
** We need to do this so that all those waiting are released.
*/
//    cpqci_fasync (-1, f, 0);

    return cpqci_close_helper(f, connectionExtension);
    //return cpqci_close_helper2(f);
}

#ifdef CPQCI_X86_64
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
static int do_cpqci_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
        int ret;
        ret = cpqci_ioctl(f->f_dentry->d_inode, f, cmd, arg);
        return ret;
}

static long cpqci_compat_ioctl(struct file *f, unsigned int cmd, unsigned long arg)
{
    switch (cmd) {
        case CPQCI_IOCTL_PHYSMEM_ALLOC:
        case CPQCI_IOCTL_PHYSMEM_FREE:
        case CPQCI_IOCTL_CREATE_CHANNEL:
        case CPQCI_IOCTL_OPEN_CHANNEL:
        case CPQCI_IOCTL_REGISTER_WAITER:
        case CPQCI_IOCTL_GET_INFO:
            return cpqci_ioctl32_handler(f, cmd, arg);
        case SEM_CREATE:
        case SEM_DESTROY:
        case SEM_UP:
        case SEM_DOWN:
        case SEM_DOWN_TIMER:
        case SEM_GET_VALUE:
        case SEM_SET_VALUE:
            return do_cpqci_ioctl(f, cmd, arg);
        default:
            return -ENOIOCTLCMD;
    }

}
int cpqci_ioctl32_handler(struct file *file, unsigned int cmd, unsigned long arg)
#else
int cpqci_ioctl32_handler(unsigned int fd, unsigned int cmd, unsigned long arg, struct file *file)
#endif
{
    int err = 0;
    int i = 0;
    mm_segment_t old_fs;
    struct _CPQCIDRVOSMEMHANDLE32 *mh32 = (struct _CPQCIDRVOSMEMHANDLE32*)arg;
    struct _CPQCIDRVIOCTLMEMHANDLE mh64;
    CPQCIDRVHANDLEPARMS32 *ub32 = (CPQCIDRVHANDLEPARMS32*) arg;
    CPQCIDRVHANDLEPARMS ub64;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
        CPQCIDRVHANDLEPARMS  __user *pub = NULL;
	struct _CPQCIDRVOSMEMHANDLE  __user *pmh = NULL;
#endif


    DBG("Entered cpqci_ioctl32_handler cmd: %p\n", cmd) ;

    switch (cmd) {
    case CPQCI_IOCTL_PHYSMEM_ALLOC:
    case CPQCI_IOCTL_PHYSMEM_FREE:
        err |= get_user(mh64.PhysicalAddress, &mh32->PhysicalAddress);
        err |= get_user(mh64.VirtualAddress, &mh32->VirtualAddress);
        err |= get_user(mh64.MemorySize, &mh32->MemorySize);
        err |= get_user(mh64.Attributes, &mh32->Attributes);
        if (err) return -EFAULT;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
	pmh = compat_alloc_user_space(sizeof(mh64));
	if (!pmh) {
		DBG("cpqci_ioctl32_handler No compat_alloc_user_space: pmh\n");
		return -EFAULT;
	}
	err |= copy_to_user(pmh, &mh64, sizeof(mh64));
	if (err) {
		DBG("CPQCI_CREATE_CHANNEL:  COMPAT COPY FAILED\n") ;
		return -EFAULT;
	}
        err = do_cpqci_ioctl(file, cmd, (unsigned long)pmh);
	if (err) {
		DBG("do_cpqci_ioctl error:  %d\n", err) ;
		return -EFAULT;
	}
	err |= copy_from_user(&mh64, pmh, sizeof(mh64));
	if (err) {
		DBG("copy_from_user error:  %d\n", err) ;
		return -EFAULT;
	}
#else
        old_fs = get_fs();
        set_fs(KERNEL_DS);
        err = sys_ioctl(fd, cmd, (unsigned long)&mh64);
        set_fs(old_fs);
#endif
        break;
    case CPQCI_IOCTL_CREATE_CHANNEL:
    case CPQCI_IOCTL_OPEN_CHANNEL:
	DBG("cpqci_ioctl32_handler:  CPQCI_IOCTL_CREATE_CHANNEL/ CPQCI_IOCTL_OPEN_CHANNEL  ENTER\n");
        for(i = 0; i < CPQCI_OBJECT_NAME_LENGTH; i++) {
            err |= get_user(ub64.ObjectName[i], &ub32->ObjectName[i]);
        }
        err |= get_user(ub64.Priority, &ub32->Priority);
        err |= get_user(ub64.SendPackets, &ub32->SendPackets);
        err |= get_user(ub64.SendPacketSize, &ub32->SendPacketSize);
        err |= get_user(ub64.RecvPackets, &ub32->RecvPackets);
        err |= get_user(ub64.RecvPacketSize, &ub32->RecvPacketSize);
        err |= get_user(ub64.KeyPresent, &ub32->KeyPresent);
        for(i = 0; i < CPQCI_KEY_SIZE; i++) {
            err |= get_user(ub64.Key[i], &ub32->Key[i]);
            err |= get_user(ub64.Ccb[i], &ub32->Ccb[i]);
        }
        err |= get_user(ub64.Error, &ub32->Error);
        err |= get_user(ub64.TimeOut, &ub32->TimeOut);
        err |= get_user(ub64.Oshandle, &ub32->Oshandle);
        if (err) return -EFAULT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)

		pub = compat_alloc_user_space(sizeof(ub64));
		if (!pub) {
			DBG("cpqci_ioctl32_handler cmd: No compat_alloc_user_space: pub\n") ;
			return -EFAULT;
		}


		err |= copy_to_user(pub, &ub64, sizeof(ub64));
		if (err) {
			DBG("CPQCI_CREATE_CHANNEL:  COMPAT COPY FAILED\n") ;
			return -EFAULT;
		}
		err = do_cpqci_ioctl(file, cmd, (unsigned long)pub);

		if (err) {
			DBG("do_cpqci_ioctl error:  %d\n", err) ;
			return -EFAULT;
		}

		err |= copy_from_user(&ub64, pub, sizeof(ub64));

		if (err) {
			DBG("copy_from_user error:  %d\n", err) ;
			return -EFAULT;
		}
#else
		old_fs = get_fs();
		set_fs(KERNEL_DS);
		err = sys_ioctl(fd, cmd, (unsigned long)&ub64);
		set_fs(old_fs);
#endif
	DBG("cpqci_ioctl32_handler:  CPQCI_IOCTL_CREATE_CHANNEL / CPQIOCTL_OPEN_CHANNEL  EXIT\n\n");
        break;
    case SEM_CREATE: 
    case SEM_DESTROY:
    case SEM_UP:
    case SEM_DOWN:
    case SEM_DOWN_TIMER:
    case SEM_GET_VALUE:
    case SEM_SET_VALUE:
    case CPQCI_IOCTL_REGISTER_WAITER:
#ifdef IOCTL_GET_INFO
    case CPQCI_IOCTL_GET_INFO:
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
        err = do_cpqci_ioctl(file, cmd, arg);
#else
        old_fs = get_fs();
        set_fs(KERNEL_DS);
        err = sys_ioctl(fd, cmd, arg);
        set_fs(old_fs);
#endif
        

break;
    }
    if(err == 0) {
        switch(cmd) {
        case CPQCI_IOCTL_PHYSMEM_ALLOC:
        case CPQCI_IOCTL_PHYSMEM_FREE:
            err |= put_user(mh64.PhysicalAddress, &mh32->PhysicalAddress);
            err |= put_user(mh64.VirtualAddress, &mh32->VirtualAddress);
            err |= put_user(mh64.MemorySize, &mh32->MemorySize);
            err |= put_user(mh64.Attributes, &mh32->Attributes);
            break;
        case CPQCI_IOCTL_CREATE_CHANNEL:
        case CPQCI_IOCTL_OPEN_CHANNEL:
            for(i = 0; i < CPQCI_OBJECT_NAME_LENGTH; i++) {
                err |= put_user(ub64.ObjectName[i], &ub32->ObjectName[i]);
            }
            err |= put_user(ub64.Priority, &ub32->Priority);
            err |= put_user(ub64.SendPackets, &ub32->SendPackets);
            err |= put_user(ub64.SendPacketSize, &ub32->SendPacketSize);
            err |= put_user(ub64.RecvPackets, &ub32->RecvPackets);
            err |= put_user(ub64.RecvPacketSize, &ub32->RecvPacketSize);
            err |= put_user(ub64.KeyPresent, &ub32->KeyPresent);
            for(i = 0; i < CPQCI_KEY_SIZE; i++) {
                err |= put_user(ub64.Key[i], &ub32->Key[i]);
                err |= put_user(ub64.Ccb[i], &ub32->Ccb[i]);
            }
            err |= put_user(ub64.Error, &ub32->Error);
            err |= put_user(ub64.TimeOut, &ub32->TimeOut);
            err |= put_user(ub64.Oshandle, &ub32->Oshandle);
            break;
        case SEM_CREATE: 
        case SEM_DESTROY:
        case SEM_UP:
        case SEM_DOWN:
        case SEM_DOWN_TIMER:
        case SEM_GET_VALUE:
        case SEM_SET_VALUE:
        case CPQCI_IOCTL_REGISTER_WAITER:
#ifdef IOCTL_GET_INFO
        case CPQCI_IOCTL_GET_INFO:
#endif
            break;
        }
    }
    DBG("cpqci_ioctl32_handler exiting... err = %d\n", err) ;
    if (err) return -EFAULT;
    return err;
}
#endif

int cpqci_ioctl(struct inode *i, struct file *f, unsigned int cmd,
        unsigned long arg)
/*++

Routine Description:

    This routine is the main dispatch routine.

Arguments:

    i - inode of open device node
    f - file handle of open device node
    cmd - operation user wants us to perform
    arg - extra data for operation

Return Value:

    The function value is the status of the operation.

--*/
{
    int status = 0;
    long req_timeout = 0;
    CPQCIADAPTEREXTENSION *adapterExtension = CpqCi;
    CPQCICONNECTIONEXTENSION *c;

    //one set of parameters
    CPQCIDRVHANDLEPARMS userbuffer;
    CPQCIDRVHANDLEPARMS*ub = &userbuffer;

    CPQCICONNECTION connection;
    CPQCICHANNEL channel;

#ifdef IOCTL_GET_INFO
    module_info_type driver_information;
#endif
    struct _CPQCIDRVIOCTLMEMHANDLE mh;

    DBG("cpqci_ioctl file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);

    switch (cmd) {
    case CPQCI_IOCTL_PHYSMEM_ALLOC:
        DBG("CPQCI_IOCTL_PHYSMEM_ALLOC Called\n");
        c = (CPQCICONNECTIONEXTENSION*) f->private_data;
        if (copy_from_user(&mh, (void*) arg, sizeof(mh))) {
	    DBG("copy from user failed %d\n", sizeof(mh));
            status =  -EFAULT;
            goto bad_ioctl_exit_0;
        }
        if (c == NULL) {
	    DBG("NULL private data\n");
            status =  -EFAULT;
            goto bad_ioctl_exit_1;
        }

        CpqCiLockAdapter();
        cpqci_new_connection_mem_handle((CPQCICONTEXT)f, c, mh.MemorySize);
        CpqCiUnlockAdapter();

        if (c->MemHandleMapped == NULL) {
	    DBG("NULL MemHandleMapped\n");
            status =  -EFAULT;
            goto bad_ioctl_exit_2;
        }

        DBG("copy_to_user(%p, %p, %x)\n", arg, c->MemHandleMapped, sizeof(mh));
        if (copy_to_user((void*) arg, c->MemHandleMapped, sizeof(mh))) {
	    DBG("copy to user failed %d\n", sizeof(mh));
            status =  -EFAULT;
            goto bad_ioctl_exit_2;
        }
        break;

    case CPQCI_IOCTL_PHYSMEM_FREE:
        DBG("CPQCI_IOCTL_PHYSMEM_FREE Called\n");
        c = (CPQCICONNECTIONEXTENSION*) f->private_data;
        if (c == NULL) {
            status =  -EFAULT;
            goto bad_ioctl_exit_0;
        }
        if (copy_from_user(&mh, (void*) arg, sizeof(mh))) {
            status =  -EFAULT;
            goto bad_ioctl_exit_0;
        }
        //Sanity check on address passed in
        if (mh.VirtualAddress != c->MemHandleMapped->VirtualAddress) {
            status =  -EFAULT;
            goto bad_ioctl_exit_1;
        }
        CpqCiLockAdapter();
        cpqci_delete_connection_mem_handle(c);
        CpqCiUnlockAdapter();
        break;

    case CPQCI_IOCTL_CREATE_CHANNEL:
	DBG("cpqci_ioctl:  CPQCI_IOCTL_CREATE_CHANNEL enter\n");
        DBG("Entered CREATE_CHANNEL ioctl %p\n", c);
        if (copy_from_user(ub, (CPQCIDRVHANDLEPARMS *) arg,
                   sizeof(CPQCIDRVHANDLEPARMS))) {
            status =  -EFAULT;
            goto bad_ioctl_exit_0;
        }
        
        c = cpqci_new_connection();
        DBG("New Connection Extension(create) %p\n", c);
        if (c == NULL) {
            status =-ENOMEM;
            goto bad_ioctl_exit_2;
        }

        if (ub->Priority == CpqCiNormalPriority) {
            if (ub->TimeOut != CPQCIWAITFOREVER) {
                status = down_timeout(&adapterExtension->ChannelSem[0],ub->TimeOut);
            } else {
                status = down_interruptible(&adapterExtension->ChannelSem[0]);
            }
        }
        else if (ub->Priority == CpqCiHighPriority) {
            if (ub->TimeOut != CPQCIWAITFOREVER) {
                status = down_timeout(&adapterExtension->ChannelSem[1],ub->TimeOut);
            } else {
                status = down_interruptible(&adapterExtension->ChannelSem[1]);
            }
        }

        if (status) {
            status =-EBADF;
            goto bad_ioctl_exit_2;
        }

        f->private_data = c;
        DBG("IOCTL_CREATE pid = %d c= %p\n", current->pid, f->private_data);
        c->Header.ObjectType = CpqCiConnection;
        if (adapterExtension->Header.ObjectType == CpqCiAdapter) {
            ub->Error = CpqCiDrvCreateChannel(f,
                adapterExtension->Adapter,
                ub->SendPackets, ub->SendPacketSize,
                ub->RecvPackets, ub->RecvPacketSize,
                (CPQCICCB *)&ub->Ccb, ub->Priority, CpqCiUserMode,
                (CPQCIKEY *)(ub->KeyPresent == CpqCiTrue ? &ub->Key : NULL),
                &connection, &channel);


            //  Construct an OS device name so that the usermode thread can obtain
            //  an OS-aware context to this connection.
            //  (this connection name isn't actually used in linux
            //   and can probably be removed)
            if (ub->Error == CpqCiSuccess) {
                c->Connection = connection;
                c->Channel = channel;
                sprintf(ub->ObjectName, "%s%u", CPQCI_CONNECTION_NAME, (unsigned int)connection);
                
                CpqCiShareContext(f, adapterExtension->Adapter, c->Connection);
//                            DbgDump( adapterExtension->Adapter, connection );
            } else
                //  If there were no more channels, clear the waiter signal
                //  for this priority queue.
            if (ub->Error == CpqCiErrNoMoreChannels) {
                //FIXME: Need logic
            } else
                //  A valid channel was removed from the free list, however, an error
                //  occured sometime after the channel was removed from the list.
                //  Since other callers could have seen an empty list during the time
                //  the channel was removed from the free list, we must signal the
                //  target priority event signal to "wake" any potential waiters if
                //  a channel was temporarily removed from the free list and replaced
                //  because the CpqCiDrvCreateChannel() routine failed.
            if (channel != CPQCI_INVALID_CHANNEL) {
                if (ub->Priority == CpqCiNormalPriority)
                    up(&adapterExtension->ChannelSem[0]);
                else if (ub->Priority == CpqCiHighPriority)
                    up(&adapterExtension->ChannelSem[1]);
            }
        } else {
            ub->Error = CpqCiErrBadObjectType;
            status = -EBADF;
            goto bad_ioctl_exit_2;
        }

        if (ub->Error == CpqCiSuccess) {
            CPQCITYPEQUAD ref;
                CpqCiOsGetRefCount(f, (CPQCITYPEQUAD*) &ref);
            DBG("*** Create Connection %d ref %d channel %d.\n", connection, ref.uint, channel);
        } else {
            DBG("CpqCiDrv!DriverDispatch: connection attempt failed.\n");
            status = -1;
            goto bad_ioctl_exit_2;
        }
        DBG("copy_to_user(%p, %p, %x)\n", arg, ub, sizeof(CPQCIDRVHANDLEPARMS));
        copy_to_user((CPQCIDRVHANDLEPARMS *) arg, ub,
                 sizeof(CPQCIDRVHANDLEPARMS));

	DBG("cpqci_ioctl:  CPQCI_IOCTL_CREATE_CHANNEL exit\n");
        break;

    /***************************************************************************
        //  This IOCTL is called to open an already existing connection between the
        //  usermode consumer and a channel.  This IOCTL can only operate on
        //  adapter objects.
        ****************************************************************************/

    case CPQCI_IOCTL_OPEN_CHANNEL:
        DBG("CPQCI_IOCTL_OPEN_CHANNEL Called\n");
        if (copy_from_user(ub, (CPQCIDRVHANDLEPARMS *) arg,
                   sizeof(CPQCIDRVHANDLEPARMS))) {
            status = -EFAULT;
            goto bad_ioctl_exit_0;
        }

        c = cpqci_new_connection();
        DBG("New Connection Extension(open) %p\n", c);
        if (c == NULL) {
            status =-ENOMEM;
            goto bad_ioctl_exit_2;
        }

        f->private_data = c;
        c->Header.ObjectType = CpqCiConnection;
        if (adapterExtension->Header.ObjectType == CpqCiAdapter) {
            ub->Error = CpqCiDrvOpenChannel(f,
                    adapterExtension->Adapter,
                    CpqCiUserMode, (CPQCICCB*)&ub->Ccb,
                    (CPQCIKEY*)&ub->Key, &connection,
                    &channel);

            //  Construct an OS device name so that the usermode thread can obtain
            //  an OS-aware context to this connection.
            //  (this connection name isn't actually used in linux
            //   and can probably be removed)
            if (ub->Error == CpqCiSuccess) {
                c->Connection = connection;
                c->Channel = channel;
                sprintf(ub->ObjectName, "%s%u", CPQCI_CONNECTION_NAME, (unsigned int)connection);

                CpqCiShareContext(f, adapterExtension->Adapter, c->Connection);
//                            DbgDump( adapterExtension->Adapter, connection );
            }
        } else {
            ub->Error = CpqCiErrBadObjectType;
            status = -EBADF;
            goto bad_ioctl_exit_2;
        }

        if (ub->Error == CpqCiSuccess) {
            CPQCITYPEQUAD ref;
                CpqCiOsGetRefCount(f, (CPQCITYPEQUAD*) &ref);
            DBG("*** Open Connection %d ref %d channel %d.\n", connection, ref.uint, channel);
        } else {
            DBG("CpqCiDrv!DriverDispatch: CpqCiDrvOpenChannel(CPQCISTATUS=%u).\n", ub->Error);
            status = -1;
            goto bad_ioctl_exit_2;
        }
        copy_to_user((CPQCIDRVHANDLEPARMS *) arg, ub,
                 sizeof(CPQCIDRVHANDLEPARMS));

        break;

    /***************************************************************************
        //  This IOCTL is called to register a thread as a waiter on the arrival
        //  of receive messages.
        ****************************************************************************/

    case CPQCI_IOCTL_REGISTER_WAITER:
        c = f->private_data;
        DBG("CPQCI_IOCTL_REGISTER_WAITER:  %p\n", c);
        DBG("ChannelEvents[%d] count (ENTRY):  %d\n",c->Channel, atomic_read(&adapterExtension->ChannelEvents[c->Channel]));

        //  We only want to register a thread already connected to the channel.
        if ((c != NULL) && (c->Header.ObjectType == CpqCiConnection)) {

            CpqCiDrvRegisterWaiter( f,
                                    adapterExtension->Adapter,
                                    c->Connection );

            status=0;
            DBG("IOCTL_REGISTER_WAITER sleep for %d ms\n", arg);
            req_timeout = wait_event_interruptible_timeout( adapterExtension->RecvSignal[c->Channel], 
                                                            ((atomic_read(&adapterExtension->ChannelEvents[c->Channel])>0)), 
                                                            MS_TO_JIFFIES(arg) );


            if (req_timeout > 0)
               atomic_dec(&adapterExtension->ChannelEvents[c->Channel]);
            else 
               DBG("wait_event_interruptible_timeout return:  %d\n", req_timeout);

            DBG("ChannelEvents[%d] count (EXIT):  %d\n",c->Channel, atomic_read(&adapterExtension->ChannelEvents[c->Channel]));

            if (signal_pending(current)) {
               DBG("Woken up abruptly\n");
               status = -EINTR;
               CpqCiDrvUnregisterWaiter (f, adapterExtension->Adapter, c->Channel);
            }

        } else {
            status = -EBADF;
            DBG("CpqCiDrv!DriverDispatch: CpqCiDrvRegisterWaiter(BADCONNECTION).\n");
            // No cleanup necessary
        }

        break;

#ifdef IOCTL_GET_INFO
    case CPQCI_IOCTL_GET_INFO:
        memset(&driver_information, 0, sizeof(module_info_type));

        strcpy(driver_information.name, "cpqci.o");
        strcpy(driver_information.revision_number, CPQEFSVER);
        strcpy(driver_information.purpose, "Compaq iLO Communication Driver for Linux");
        
        driver_information.revision_date.year = BUILD_YEAR;
        driver_information.revision_date.month= BUILD_MONTH;
        driver_information.revision_date.day  = BUILD_DAY;

        if (arg) copy_to_user((module_info_type*) arg, &driver_information, sizeof(module_info_type));

        break;
#endif
    /***************************************************************************
        //  The specified IOCTL is not supported.
        ****************************************************************************/

    default:
        DBG("default ioctl %x\n", cmd);
        //First check for sem ioctl
        if (cmd >= SEM_CREATE && cmd <= SEM_SET_VALUE) {
            status = sem_ioctl(i, f, cmd, arg);
        } else {
            status = -ENOIOCTLCMD;
        }
    }

    DBG("IOCTL cmd %x arg %p returns %d\n", cmd, arg, status);
    return status;


bad_ioctl_exit_2:
    if (c) { 
        DBG("(connection)cpqci_delete_connection(%p)\n", c);
        CpqCiLockAdapter(); 
        cpqci_delete_connection(c);
        f->private_data = NULL; //c ??
        CpqCiUnlockAdapter();
    }
bad_ioctl_exit_1:
    if (arg && ub) {
        copy_to_user((CPQCIDRVHANDLEPARMS *) arg, ub,
             sizeof(CPQCIDRVHANDLEPARMS));
    }

bad_ioctl_exit_0:
    return status;
}

/*
** cpqci_fasync
**
** This is a standard Linux routine that is used in conjunction 
** with cpqci_poll to allow an application to wait on an event.
*/
int cpqci_fasync(int fd, struct file *f, int on)
{
    CPQCIADAPTEREXTENSION *adapterExtension = CpqCi;
    CPQCICONNECTIONEXTENSION *priv = f->private_data;
        int                      result;

        result = fasync_helper(fd, f, on, &priv->fasync_queue);

        return (result);
}




/*
** cpqci_mmap
**
** Remaps a physical page range into the virtual range specified by vma.
** The only way we should ever get into this function is as a result of
** a call to do_mmap in CpqCiDrvOsMap.  Any other call (e.g. from mmap
** in user space) will fail.
*/
int cpqci_mmap(struct file *f, struct vm_area_struct *vma)
{
    unsigned long size = vma->vm_end - vma->vm_start;
    CPQCICONNECTIONEXTENSION *e = f->private_data;
    unsigned long pa;
    unsigned int nopage_method = 0;
    unsigned pcimem;
    int rc;

    DBG("cpqci_mmap file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);

    if (e == NULL) return -EINVAL;

/*
** If the daddr is equal to the paddr passed in via the "Context" then
** this is the DMA mapped memory.  Otherwise, this is most likely the 
** MMIO memory for the PCI Device (via the BAR). We must make this
** distinction because of a GART being used as an IOMMU.
*/
    pcimem = (e->daddr == e->paddr) ? 0 : 1;
    if (e->daddr == e->paddr) {
       e->paddr = virt_to_phys(e->vaddr);  /* if needed for cpqci_vma_nopage_regular */
       DBG("cpqci_mmap:  Detected IOMMU:  daddr (%p) == paddr (%p)\n", e->daddr, e->paddr);
    }

    pa = e->paddr;

#ifdef THIS_IS_NOT_NEEDED
    if (pa < (unsigned long)pci_mem_start) {
        DBG("Mapping regular memory at %p (high mem %p)\n", pa, pci_mem_start);
        nopage_method = 0;
    } else {       
        DBG("Mapping PCI memory at %p (high mem %p)\n", pa, pci_mem_start);
        nopage_method = 0;
    }
#endif

    if ((e->MemHandleMalloc != NULL) && (pa == e->MemHandleMalloc->PhysicalAddress)) {
        DBG("This is the mapping of the additional memory buffer with phys %p ...\n", e->MemHandleMalloc->PhysicalAddress);
        nopage_method = 0;
    }



    vma->vm_flags |= VM_RESERVED;
    vma->vm_private_data= e;
    DBG("Manipulating vma->vm_private_data = %p\n", vma->vm_private_data);


    if (nopage_method == 0) {
        vma->vm_flags |= VM_IO;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16)
                DBG("remap 0x%x from 0x%x:0x%x\n", vma->vm_start, pa, size);
                if (pcimem) {
		    vma->vm_flags |= VM_PFNMAP;
                    rc = io_remap_pfn_range(vma, vma->vm_start, (pa)>>PAGE_SHIFT, size, vma->vm_page_prot);
                } else {
                    rc = remap_pfn_range(vma, vma->vm_start, (pa)>>PAGE_SHIFT, size, vma->vm_page_prot);
                }
                if (rc) {
                    DBG("pcimem = %d, remap rc = %d\n", pcimem, rc);    
                    return -EAGAIN;
                }
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
		DBG("CALLING:  remap_page_range (2.6.9) . . .\n");
                if (remap_page_range(vma, vma->vm_start, pa, size, vma->vm_page_prot))
                                        return -EAGAIN;

#else
		DBG("CALLING:  remap_pfn_range . . .\n");
                if (remap_pfn_range(vma, vma->vm_start, (pa)>>PAGE_SHIFT, size, vma->vm_page_prot))
                            return -EAGAIN;
#endif
            vma->vm_ops = &cpqci_vm_ops_high;
    }

    if (nopage_method == 1) {
        vma->vm_flags &= ~VM_IO;
            vma->vm_ops = &cpqci_vm_ops_regular;
    }

    DBG("Protection bits %p\n", vma->vm_page_prot);
    DBG("Flag bits %p\n", vma->vm_flags);

    return 0;
}

/*
** cpqci_intr
**
** Interrupt handler for the channel interface.  This function gets
** a list of channels with pending messages and wakes up any processes
** waiting on those channels.
*/
irqreturn_t cpqci_intr(int irq, void *dev_id, struct pt_regs *regs)
{
    CPQCICHANNEL channel;
    CPQCICHANNELS channelList = 0;
    CPQCICHANNELS deferredList = 0;
    CPQCIADAPTEREXTENSION *adapterExtension = (CPQCIADAPTEREXTENSION *)dev_id;
    int i;
    int rc;
    irqreturn_t retVal = IRQ_NONE;    

    //This is to ensure that this is really our interrupt
    if (adapterExtension == CpqCi) {
       if ( (CpqCiDrvGetPendingChannels(NULL, adapterExtension->Adapter, &channel, &channelList, &deferredList)) == CpqCiSuccess ) {

          DBG("Interrupt\n");
          retVal = IRQ_HANDLED;

          //  Spin down the pending channel list until there are no more
          //  pending channels.  This has the effect of waking up waiters
          //  on the respective channel receive FIFO.
    
          do {

            DBG("Intr Channel = %d\n", channel);
            DBG("ChannelEvents[%d] count (INTR):  %d\n",channel, atomic_read(&adapterExtension->ChannelEvents[channel]));
            atomic_inc(&adapterExtension->ChannelEvents[channel]);
            if (CpqCiDrvUnregisterWaiter (NULL, adapterExtension->Adapter, channel) == CpqCiSuccess) {

// This is correct for Windows but I am not sure for Linux
// Need STRESS TEST to prove this out.  Currently, this causes
// ChannelEvents to get incremented 2x for each interrupt.
#ifdef PMS_NEED_STRESS
                deferredList |= (1 << channel);  // THIS IS CORRECT!  DO NOT TOUCH!
#endif
                wake_up_all(&adapterExtension->RecvSignal[channel]);
            }

          } while ( (CpqCiDrvGetPendingChannels(NULL, adapterExtension->Adapter, &channel, &channelList, &deferredList)) == CpqCiSuccess );

       } else {
          DBG("CpqCiDrvGetPendingChannels returned an ERROR\n");
       }

    }

    return (retVal);
}


//****************************************************************
//*         VMA OPERATIONS FOR SHARED MEMORY.
//****************************************************************
//
static unsigned long cpqci_do_mmap(CPQCICONTEXT Context, CPQCIPHYSICAL PhysicalAddress, CPQCISIZE MappedSize, unsigned long prot)
{
    unsigned long va;

    struct file *f = (struct file *) Context;
    CPQCICONNECTIONEXTENSION *e = f->private_data;
    DBG("cpqci_do_mmap file pointer %p refcount %d\n", f, f ? atomic_read((atomic_t*)&f->f_count):0);

    e->paddr = PhysicalAddress;

    DBG("cpqci_do_mmap(%p, %d)\n", PhysicalAddress, MappedSize);
    down_write(&current->mm->mmap_sem);
    va = do_mmap(f, 0, MappedSize, prot, MAP_SHARED, 0);
    up_write(&current->mm->mmap_sem);
    DBG("cpqci_do_mmap VA returned: %p\n", va);
    //The high number would indicate a failure to map properly
    if (va > 0xfffffc00) va = 0;
    return va;
}

static void cpqci_do_munmap(CPQCICONTEXT Context, CPQCIPHYSICAL VirtualAddress, CPQCISIZE MappedSize)
{
            DBG("munmap(%p, %d)\n", VirtualAddress, MappedSize);
#if defined(MUNMAP4)
            if (current->mm) {
                down_write(&current->mm->mmap_sem);
                do_munmap(current->mm, (unsigned long) VirtualAddress, MappedSize, 1);
                up_write(&current->mm->mmap_sem);
            }
#else
            if (current->mm) {
                down_write(&current->mm->mmap_sem);
                do_munmap(current->mm, (unsigned long) VirtualAddress, MappedSize);
                up_write(&current->mm->mmap_sem);
            }
#endif
}

// **********************************************************************
//VMA operations
// **********************************************************************

/*
 * Compatibility  
 */
static void* vma_private_data(struct vm_area_struct *vma)
{
    return vma->vm_private_data;
}


static void cpqci_vma_open(struct vm_area_struct *vma)
{ 
    DBG("cpqci_vma_open vm_start=%p (private %p) \n", vma->vm_start, vma->vm_private_data);
}

static void cpqci_vma_close(struct vm_area_struct *vma)
{ 
    DBG("cpqci_vma_close vm_start=%p (private %p) \n", vma->vm_start, vma->vm_private_data);
}


static struct page* cpqci_vma_nopage_regular(struct vm_area_struct *vma, unsigned long address, int *write) 
{
    unsigned long offset;
    struct page* page = NOPAGE_SIGBUS;
    CPQCICONNECTIONEXTENSION* e = (CPQCICONNECTIONEXTENSION*) vma_private_data(vma);
    CPQCICONNECTION connection ;
        struct _CPQCIADAPTEROBJECT *pao;
    CPQCIDRVOSMEMHANDLE mh;
    unsigned long va;
//FIXME: critical section ??
    //
    DBG("nopage: vma_start= %p\n", vma->vm_start);
    DBG("nopage: vma_end= %p\n", vma->vm_end);
    DBG("nopage: address = %p\n", address);
    offset = (address - vma->vm_start);
    DBG("nopage: offset = %p\n", offset);
    if (offset >= (vma->vm_end-vma->vm_start)) goto nopage_regular_out;

    if (e == NULL) goto nopage_regular_out;

    connection = e->Connection;
        pao = CpqCi->Adapter;
    mh = (CPQCIDRVOSMEMHANDLE) GetMemoryHandle(pao, connection);
    va = (unsigned long)__va(mh->PhysicalAddress + offset);

#ifdef DEBUG
    DBG("nopage: va = %p\n", va);
    for (i=0; i<16; i++) {
        DBG("%2.2x ", ((unsigned char*)va)[i]);
    }
    DBG("\n");
#endif
    page = virt_to_page(va);
    DBG("nopage: page = %p\n", page);
    get_page(page);

nopage_regular_out:

    return page;

}

static struct page* cpqci_vma_nopage_high(struct vm_area_struct *vma, unsigned long address, int *write) 
{
    struct page* page = NOPAGE_SIGBUS;
    DBG("nopage: vma_start= %p\n", vma->vm_start);
    DBG("nopage: vma_end= %p\n", vma->vm_end);
    DBG("nopage: address = %p\n", address);
    return page;
}


//****************************************************************
//*         DRIVER MEMORY MANAGEMENT ROUTINES.
//****************************************************************

#define _cached         (0 << 0)
#define _uncached       (1 << 0)

#define _kernelmode     (0 << 1)
#define _usermode       (1 << 1)

#define _allocated      (0 << 2)
#define _mapped         (1 << 2)

#define _noncontiguous  (0 << 3)
#define _contiguous     (1 << 3)

#define _memoryobj      (0 << 4)   
#define _dmaobj         (1 << 4)   

#define _objecttype     (3 << 5)    // 2-bits of info

CPQCIDRVOSHANDLE CpqCiDrvOsMallocContiguous(CPQCICONTEXT Context,
                        CPQCISIZE AllocationSize)
/*++

Routine Description:
    This routine allocates physically contiguous memory for DMA

Arguments:
    AllocationSize - The size in bytes of the memory to allocate.
    
Return Value:
    A handle representing the memory region allocated.  A value of NULL (0)
    indicates no more resources to fulfill the request.

Caveats:
    The allocated memory MUST have the following attributes:

        CACHED
        NON-PAGED
        CONTIGUOUS

--*/
{
    CPQCIDRVOSMEMHANDLE mh = NULL;
    CPQCIADAPTEREXTENSION *ah = CpqCi;
    struct file* f = (struct file*) Context;
    CPQCICONNECTIONEXTENSION *e = f->private_data;
    void *va = NULL;
    int i;

#ifdef SKIPTHIS
#if !defined(CPQCI_X86_64) && LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,10)
   ah->pdev->dev.dma_mem = NULL;
   *ah->pdev->dev.dma_mask = 0;
#endif
#endif


    //  Make sure the requested allocation size is valid.
    if (AllocationSize) {
        mh = kmalloc(sizeof(struct _CPQCIDRVOSMEMHANDLE), GFP_KERNEL);
	DBG("(CpqCiDrvOsMallocContiguous) mh is :  %p\n", mh);

        if (mh) {
            memset(mh, 0, (sizeof(struct _CPQCIDRVOSMEMHANDLE)));
            DBG("Before dma_alloc_coherent - AllocationSize:  0x%x\n",AllocationSize);
            va = dma_alloc_coherent( &ah->pdev->dev, AllocationSize, 
                                     &mh->dma_handle, GFP_KERNEL );

//            va = dma_alloc_coherent( NULL, AllocationSize, 
//                                     &mh->dma_handle, GFP_KERNEL );
            if (va) {
                mh->MemorySize = AllocationSize;
                mh->PhysicalAddress = mh->dma_handle;
                mh->VirtualAddress = va;
                mh->Attributes =
                    _dmaobj | _cached | _kernelmode |
                    _allocated | _contiguous;
		DBG("dma_alloc_coherent:  va: 0x%x  pa: 0x%x  size: 0x%x \n", va, mh->dma_handle, (unsigned long)AllocationSize);

                for (i = 0; i < AllocationSize; i+=PAGE_SIZE) {
                    SetPageReserved(virt_to_page(va + i));
                }

            /*
            ** We need to save off ALL of the information returned from
            ** dma_alloc_coherent() so we can determine which method to
            ** use to mmap() pages to the user.
            */
               e->daddr = mh->dma_handle;
               e->vaddr = va;

            } else {
                printk(KERN_ERR "CpqCiDrvOsMallocContiguous: dma_alloc_coherent failed for %d bytes\n", (int)AllocationSize);
                kfree(mh);
                mh = NULL;
            }

        } else {
            printk(KERN_ERR "CpqCiDrvOsMallocContiguous: Out of Memory!!\n");
            return (NULL);
        }

    } else {
       DBG("CpqCiDrvOsMallocContiguous:  MappedSize is zero\n");
       return (NULL);
    }

    return mh;
}

CPQCIDRVOSHANDLE CpqCiDrvOsMalloc(CPQCICONTEXT Context,
                  CPQCISIZE AllocationSize)
/*++

Routine Description:
    This routine allocates noncontiguous memory for the requested size.

Arguments:
    AllocationSize - The size in bytes of the memory to allocate.
    
Return Value:
    A handle representing the memory region allocated.  A value of NULL (0)
    indicates no more resources to fulfill the request.

Caveats:
    The allocated memory MUST have the following attributes:

        CACHED
        NON-PAGED
        CONTIGUOUS

--*/
{
    CPQCIDRVOSMEMHANDLE mh = NULL;
    void *va = NULL;

    //  Make sure the requested allocation size is valid.
    if (AllocationSize) {
        mh = kmalloc(sizeof(struct _CPQCIDRVOSMEMHANDLE), GFP_KERNEL);
	DBG("(CpqCiDrvOsMalloc) mh = %p\n", mh);
        if (mh) {
            //  Allocate the actual memory from non-paged pool, physically contiguous.
            DBG("Malloc(%x) = ", AllocationSize);
            va = kmalloc(AllocationSize, GFP_KERNEL);
            DBG("%p\n", va);
	    DBG("(CpqCiDrvOsMalloc) va = %p\n", va);
            if (va) {
                //This is to prevent junk pointers, etc. 
                memset(va, 0, AllocationSize);
                //  Update the memory handle.
                mh->PhysicalAddress = __pa(va);
                mh->VirtualAddress = va;
                mh->MemorySize = AllocationSize;
                mh->Attributes =
                    _memoryobj | _cached | _kernelmode |
                    _allocated | _contiguous;
            } else {
                printk(KERN_ERR "CpqCiDrvOsMalloc: Out of Memory - NO VA!\n");
                kfree(mh);
                mh = NULL;
            }

        } else {
            printk(KERN_ERR "CpqCiDrvOsMalloc: Out of Memory!!\n");
            return (NULL);
        }

    } else {
       DBG("CpqCiDrvOsMalloc:  MappedSize is zero\n");
       return (NULL);
    }

    return mh;
}

void CpqCiDrvOsFree(CPQCICONTEXT Context, CPQCIDRVOSHANDLE MemoryHandle)
{
    CPQCIDRVOSMEMHANDLE mh = MemoryHandle;
    CPQCIADAPTEREXTENSION *ah = CpqCi;

    int i;

    //mh needs to be defined, otherwise noop
    if (!mh) return;

    if (mh->VirtualAddress != 0) {
        if (mh->Attributes & _mapped) {
            if (mh->Attributes & _usermode) {
                DBG( "CpqCiDrvOsFree: Calling cpqci_do_munmap(%p, %d)\n", 
                      mh->VirtualAddress, mh->MemorySize );

                cpqci_do_munmap(Context, (unsigned long) mh->VirtualAddress, mh->MemorySize);

            } else {   /* kernel mode allocation */

                DBG( "CpqCiDrvOsFree: Calling iounmap(%p, %d)\n", 
                      mh->VirtualAddress, mh->MemorySize );

                iounmap(mh->VirtualAddress);
                release_mem_region(mh->PhysicalAddress, mh->MemorySize);
            }

        } else if (mh->Attributes & _dmaobj) {

                if (mh->VirtualAddress != NULL) {
                   for (i = 0; i < mh->MemorySize; i += PAGE_SIZE) {
//                      clear_bit(PG_locked, &virt_to_page(mh->VirtualAddress + i)->flags);
//                      atomic_dec(&virt_to_page(mh->VirtualAddress + i)->_count);
                      ClearPageReserved(virt_to_page(mh->VirtualAddress + i));
                   }

                   DBG( "CpqCiDrvOsFree: dma_free_coherent(%p, 0x%x) dma_handle:  %p\n\n", 
                         mh->VirtualAddress, mh->MemorySize, mh->dma_handle );

                   dma_free_coherent( &ah->pdev->dev, mh->MemorySize,
                                      mh->VirtualAddress, mh->dma_handle );
                } else {
                   printk(KERN_ERR "CpqCiDrvOsFree: mh->VirtualAddress == NULL\n");
                }

        } else {  /* regular memory object */
           DBG( "CpqCiDrvOsFree: kfree(%p, %d)\n", mh->VirtualAddress, mh->MemorySize );
           kfree(mh->VirtualAddress);
        }
    }

    if (mh) {
	DBG("(CpqCiDrvOsFree) Freeing  mh: %p\n",mh);
        kfree(mh);
    }

    return;
}

CPQCIDRVOSHANDLE
CpqCiDrvOsDupHandle(CPQCICONTEXT Context, CPQCIDRVOSHANDLE Handle)
{
    CPQCIDRVOSHANDLE mh;

    if (!Handle) return NULL;

    mh = kmalloc(sizeof(struct _CPQCIDRVOSMEMHANDLE), GFP_KERNEL);
    DBG("(OSMEMHANDLE)Kmalloc(%d) = %p\n", sizeof(struct _CPQCIDRVOSMEMHANDLE), mh);
    if (!mh) return mh;

    memcpy(mh, Handle, sizeof(struct _CPQCIDRVOSMEMHANDLE));
    return mh;
}

CPQCIDRVOSHANDLE
CpqCiDrvOsMap(CPQCICONTEXT Context, CPQCIPHYSICAL PhysicalAddress,
          CPQCISIZE MappedSize,
          CPQCIMODE AddressMode, CPQCICACHETYPE CacheType)
{
    CPQCIDRVOSMEMHANDLE mh = NULL;
    void *va = NULL;

    //  Make sure the requested allocation size is valid.
    if (MappedSize) {
        //  Allocate a memory handle from non-paged pool.
        mh = kmalloc(sizeof(struct _CPQCIDRVOSMEMHANDLE), GFP_KERNEL);
	DBG("(CpqCiDrvOsMap)  mh = %p\n", mh);
        if (mh) {
            if (AddressMode == CpqCiKernelMode) {

               if (request_mem_region(PhysicalAddress, MappedSize, DEVICE_NAME) == NULL) {
                  printk(KERN_ERR "CpqCiDrvOsMap: request_mem_region for 0x%x (size %d) failed!\n", PhysicalAddress, (int)MappedSize);
                  kfree(mh);
                  return (NULL);
               }


               if (CacheType == CpqCiCached) {
                  DBG("ioremap(%p, %d) = ", PhysicalAddress, MappedSize);
                  va = ioremap(PhysicalAddress, MappedSize);
                  DBG("%p\n", va);
               } else {
                  DBG("ioremap_nocache(%p, %d) = ", PhysicalAddress, MappedSize);
                  va = ioremap_nocache(PhysicalAddress, MappedSize);
                  DBG("%p\n", va);
               }

            } else {
                //
                // This is user mode space
                //
                va = (void*) cpqci_do_mmap(Context, PhysicalAddress, MappedSize, PROT_READ | PROT_WRITE);
                DBG("cpqci_do_mmap(%p, %d) = ", PhysicalAddress, MappedSize);
                DBG("%p\n", va);
            }

        } else {
           //
           // No Memory available for MemoryHandle (mh)
           //
           printk(KERN_ERR "CpqCiDrvOsMap: Out of Memory!!\n\n");
           return (NULL);
        }

    } else {
       DBG("CpqCiDrvOsMap:  MappedSize is zero\n");
       return (NULL);
    }

    //  If a valid va was obtained, fill in the memory handle.
    if (va) {
        mh->PhysicalAddress = PhysicalAddress;
        mh->VirtualAddress = va;
        mh->MemorySize = MappedSize;
        mh->Attributes = _memoryobj | _mapped;
        mh->Attributes |=
            ((AddressMode == CpqCiUserMode) ? _usermode :
             _kernelmode);
        mh->Attributes |=
            ((CacheType == CpqCiCached) ? _cached : _uncached);
    } else if (mh) {
        printk(KERN_ERR "CpqCiDrvOsMap: failed to map 0x%x\n", PhysicalAddress);
        kfree(mh);
        mh = NULL;
    }

    return mh;
}

CPQCIVIRTUAL CpqCiDrvOsGetVA(CPQCICONTEXT Context,
                 CPQCIDRVOSHANDLE MemoryHandle)
{
    if (!MemoryHandle) {
        DBG("CpqCiDrvOSGetVA: MemoryHandle == NULL\n");
        return 0;
    }
    DBG( "CpqCiDrvOSGetVA: Physical Address == %p\n", 
         ((CPQCIDRVOSMEMHANDLE)MemoryHandle)->PhysicalAddress );

    DBG( "CpqCiDrvOSGetVA: Virtual Address == %p\n", 
         ((CPQCIDRVOSMEMHANDLE)MemoryHandle)->VirtualAddress );

    return (CPQCIVIRTUAL) ((CPQCIDRVOSMEMHANDLE)
                   (MemoryHandle))->VirtualAddress;
}

CPQCIPHYSICAL CpqCiDrvOsGetPA(CPQCICONTEXT Context,
                  CPQCIDRVOSHANDLE MemoryHandle)
{
    if (!MemoryHandle) {
        DBG("CpqCiDrvOSGetVA: MemoryHandle == NULL\n");
        return 0;
    }

    DBG( "CpqCiDrvOSGetPA: Physical Address == %p\n", 
         ((CPQCIDRVOSMEMHANDLE)MemoryHandle)->PhysicalAddress );

#ifdef SNAFU
    return *((CPQCIPHYSICAL *)
         (&((CPQCIDRVOSMEMHANDLE) (MemoryHandle))->PhysicalAddress));
#endif

    return (CPQCIPHYSICAL)(((CPQCIDRVOSMEMHANDLE)(MemoryHandle))->PhysicalAddress);
}

CPQCISIZE CpqCiDrvOsGetSize(CPQCICONTEXT Context,
                CPQCIDRVOSHANDLE MemoryHandle)
{
    if (!MemoryHandle) {
        DBG("CpqCiDrvOSGetVA: MemoryHandle == NULL\n");
        return 0;
    }
    return (CPQCISIZE) ((CPQCIDRVOSMEMHANDLE)
                (MemoryHandle))->MemorySize;
}

CPQCICACHETYPE CpqCiDrvOsGetCacheType(CPQCICONTEXT Context,
                      CPQCIDRVOSHANDLE MemoryHandle)
{
    if (!MemoryHandle) {
        DBG("CpqCiDrvOSGetVA: MemoryHandle == NULL\n");
        return 0;
    }
    return (((CPQCIDRVOSMEMHANDLE) (MemoryHandle))->Attributes &
        _uncached) ? CpqCiUncached : CpqCiCached;
}

CPQCIMODE CpqCiDrvOsGetMode(CPQCICONTEXT Context,
                CPQCIDRVOSHANDLE MemoryHandle)
{
    if (!MemoryHandle) {
        DBG("CpqCiDrvOSGetVA: MemoryHandle == NULL\n");
        return 0;
    }
    return (((CPQCIDRVOSMEMHANDLE) (MemoryHandle))->Attributes &
        _usermode) ? CpqCiUserMode : CpqCiKernelMode;
}


CPQCIVOID CpqCiDrvOsStall(  CPQCICONTEXT Context, 
                CPQCITIMEOUT MicroSeconds)
{
   unsigned long ulDelay;

   if (MicroSeconds < 1000) {     /* one millisecond       */
      udelay (MicroSeconds);

   } else {
      if (MicroSeconds < 10000) {   /* Slightly longer wait  */
         ulDelay = MicroSeconds / 1000;          /* convert to mSec       */
      } else {
         ulDelay = 10000;
      }

      mdelay (ulDelay);
   }

}

CPQCISTATUS
CpqCiOsGetRefCount(
    CPQCICONTEXT Context, 
    CPQCITYPEQUAD* RefCount
   )
{
    struct file* fp = (struct file*) Context;
    
    memset(RefCount, 0, sizeof(CPQCITYPEQUAD));

    DBG("RefCount of File %p\n", fp);

    if (!fp) return CpqCiSuccess;

    RefCount->uint = atomic_read((atomic_t*)&fp->f_count);
    DBG("pid %d: RefCount %p\n", current->pid, RefCount->uint);

    return CpqCiSuccess;
}

CPQCIVIRTUAL
CpqCiMemSet(
    CPQCIVIRTUAL dest, 
    u8_t t, 
    CPQCISIZE s) 
{
    return memset(dest, t, s);
}


CPQCISIZE CpqCiSPrintf(
    CPQCICHAR* buf, 
    const CPQCICHAR* fmt, 
    ...
    ) 
{
    va_list args;
    int i;
    va_start(args, fmt);
    i=vsprintf(buf, fmt, args);
    va_end(args);
    return i;
}

void* 
CpqCiGetCurrentTask(void)
{
    void* curr = (void*) current;
    return curr;
}

CPQCISIZE CpqCiSNPrintf(
        CPQCICHAR* buf,
        CPQCISIZE size,
        const CPQCICHAR* fmt,
        ...
        )
{
        va_list args;
        int i;
        va_start(args, fmt);
        i=vsnprintf(buf, size, fmt, args);
        va_end(args);
        return i;
}



static int __devinit hpqci_pci_probe(struct pci_dev *pdev,
                    const struct pci_device_id *ent)
{
    int				status = 0;
    CPQCIADAPTEREXTENSION	*adapterExtension;


    //*******************************************************************
    //  Create the master adapter device object used to track access to
    //  this driver.  The master object does not contain a connection, 
    //  only an adapter object.
    //*******************************************************************

    adapterExtension = kmalloc(sizeof(CPQCIADAPTEREXTENSION), GFP_KERNEL);
    DBG("(hpqci_pci_probe)kmalloc(%d) = %p\n", sizeof(CPQCIADAPTEREXTENSION), adapterExtension);
    if (!adapterExtension)
        return (ENOMEM);

    memset(adapterExtension, 0, sizeof(CPQCIADAPTEREXTENSION));

    CpqCi = adapterExtension;  // Set the global pointer.

    status = pci_enable_device(pdev);
    if (status) {
        printk(KERN_ERR "hpqci_pci_probe: %s: couldn't enable PCI device\n",
               pci_name(pdev));
        kfree(adapterExtension);
        CpqCi = NULL;
        return (status);
    }

    adapterExtension->pdev = pdev;
    return (status);
}




static void __devexit hpqci_pci_shutdown(struct pci_dev *pdev)
{
    int				status = 0;
    CPQCIADAPTEREXTENSION	*adapterExtension;

    DBG("hpqci_pci_shutdown called\n");
    if (CpqCi) {
	DBG("(hpqci_pci_shutdown) kfree CpqCi = %p\n",CpqCi);
        kfree(CpqCi);
        CpqCi = NULL;
    }

}



static __devinit int init_hpqci(void)
/*++

Routine Description:

    This is the initialization routine for this driver.

Arguments:

Return Value:

    The function value is the final status from 
    the initialization operation.

--*/
{
    int status;
    int ret = 0;
    int register_status;
    int irq_status;
    struct pci_dev *pdev;

    CPQCICONNECTIONEXTENSION    *connectionExtension;
    CPQCIADAPTEREXTENSION       *adapterExtension;
    CPQCICONNECTIONS            connections = 256; //FIXME: How many do we need?
    struct _CPQCIADAPTEROBJECT  *adapter;
    CPQCISTATUS                 error;

    CPQCIPHYSICAL               mmioreg = 0x0;
    CPQCISIZE                   mmioregsize = 0;

    CPQCICHANNEL                channel;

    //*******************************************************************
    //  Set default values.
    //*******************************************************************

    status              = 0;
    irq_status          = -1;
    register_status     = -1;
    adapterExtension    = NULL;
    connectionExtension = NULL;
    error               = CpqCiSuccess;
    CpqCi = adapter     = NULL;


#if defined(CPQCI_X86_64) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10)
	
	ret = register_ioctl32_conversion(CPQCI_IOCTL_PHYSMEM_ALLOC, cpqci_ioctl32_handler);
	ret |= register_ioctl32_conversion(CPQCI_IOCTL_PHYSMEM_FREE, cpqci_ioctl32_handler);
	ret |= register_ioctl32_conversion(CPQCI_IOCTL_CREATE_CHANNEL, cpqci_ioctl32_handler);
	ret |= register_ioctl32_conversion(CPQCI_IOCTL_OPEN_CHANNEL, cpqci_ioctl32_handler);
	ret |= register_ioctl32_conversion(CPQCI_IOCTL_REGISTER_WAITER, cpqci_ioctl32_handler);
	ret |= register_ioctl32_conversion(CPQCI_IOCTL_GET_INFO, cpqci_ioctl32_handler);

	ret |= register_ioctl32_conversion(SEM_CREATE, NULL);
	ret |= register_ioctl32_conversion(SEM_DESTROY, NULL);
	ret |= register_ioctl32_conversion(SEM_UP, NULL);
	ret |= register_ioctl32_conversion(SEM_DOWN, NULL);
	ret |= register_ioctl32_conversion(SEM_DOWN_TIMER, NULL);
	ret |= register_ioctl32_conversion(SEM_GET_VALUE, NULL);
	ret |= register_ioctl32_conversion(SEM_SET_VALUE, NULL);

	if (ret) {
		printk(KERN_WARNING "cpqci:  Error registering ioctl32\n");
        	goto bad_exit;
        }
#endif


#ifdef PROBABLY_NOT_NEEDED
    //*******************************************************************
    //  Register the Device Driver
    //*******************************************************************

    status = driver_register(&hpqci_base_driver);
    if (status) {
        printk(KERN_ERR "init_hpqci: Unable to register driver: status (%d)\n", status);
        return status;
    }

#endif

    //*******************************************************************
    //  Register for PCI Device
    //*******************************************************************

//    status = pci_module_init(&hpqci_pci_driver);
    status = pci_register_driver(&hpqci_pci_driver);
    if (status < 0) {
        printk(KERN_ERR "init_hpqci: Unable to initial PCI interface: status (%d)\n", status);
        return -ENODEV;;
    } else {
	status = 0;   /* Gets set to "1" in older version of pci_register_driver */
    }


    adapterExtension = CpqCi;
    pdev = CpqCi->pdev;
    mmioreg = (CPQCIPHYSICAL)pci_resource_start(pdev, 1); // BAR 1, Offset 0x14
    mmioregsize = (CPQCISIZE)pci_resource_len(pdev, 1);

    //*******************************************************************
    //  Update the adapter object's device extension information.
    //*******************************************************************

    adapterExtension->Header.ObjectType = CpqCiAdapter;
    adapterExtension->Header.AdapterExtension = adapterExtension;
    adapterExtension->ConnectionList = NULL;

    //*******************************************************************
    // Initialize Linux Specific support before creating adapter in the
    // shared code.
    //*******************************************************************

    sema_init(&adapterExtension->ChannelSem[0], 20);
    sema_init(&adapterExtension->ChannelSem[1], 4);
    sema_init(&connection_lock, 1);
    spin_lock_init(&adapterExtension->AdapterLock);

    for (channel = 0; channel < CPQCI_MAX_CHANNELS; channel++) {
        init_waitqueue_head(&adapterExtension->RecvSignal[channel]);
        spin_lock_init(&adapterExtension->ChannelLock[channel]);
    }
    
    //*******************************************************************
    //  Create a device specific adapter object.
    //*******************************************************************

    error = CpqCiDrvCreateAdapter(NULL, mmioreg, mmioregsize, connections, &adapter);
    if (error != CpqCiSuccess) {
        DBG("CpqCiDrv!init_module: CpqCiDrvCreateAdapter(CPQCISTATUS=%u)\n", error);
        status = ENODEV;
        goto bad_exit;
    }

    //*******************************************************************
    // Complete initialization of Adapter Extension
    //*******************************************************************

    adapterExtension->Adapter = adapter;
    adapterExtension->ConnectionList = NULL;


    register_status = misc_register(&cpqci_miscdev);

    DBG("adapterExtension->pdev->irq:  %d\n", adapterExtension->pdev->irq);
    irq_status = request_irq( adapterExtension->pdev->irq, 
                              (irqreturn_t (*)(int, void*, struct pt_regs *))cpqci_intr, 
                              (IRQF_DISABLED | IRQF_SHARED), 
                              DEVICE_NAME, 
                              CpqCi );

#if defined(CONFIG_PROC_FS)
    cpqci_entry_print_cpqci = create_proc_entry("hpqci", S_IFREG | S_IRUGO, &proc_root);
    if (cpqci_entry_print_cpqci) {
        cpqci_entry_print_cpqci->read_proc = &cpqci_read_print_cpqci;
    } else {
        printk(KERN_ERR "proc entry creation 'hpqci' FAILED.\n");
        goto bad_exit;
    }
#endif

    return status;

bad_exit:

    cleanup_module_helper(irq_status, (pdev == NULL) ? 0 : pdev->irq, register_status, adapter, adapterExtension);
    return status;

}



static __exit void cleanup_hpqci(void)
{

    cleanup_module_helper(0, CpqCi->pdev->irq, 0, CpqCi->Adapter, CpqCi);
    pci_unregister_driver(&hpqci_pci_driver);

    CpqCi = NULL;
}

module_init(init_hpqci);
module_exit(cleanup_hpqci);

