/*++

(C) Copyright 2000-2006 Hewlett-Packard Development Company, L.P.

Module Name:

    CpqCilx.c

Abstract:

    This module contains the host algorithms used to enqueue
    and dequeue packets to/from a channel.

Author:

    Thomas J. Bonola (Tom.Bonola@compaq.com) Apr-2000

Environment:

    OS Agnostic
    Processor mode Agnostic.

Compiler Settings:

    This module requires the following compiler settings.

        *   1-byte packed data structures
        *   All functions must adhere to standard 'C' calling convention
        
Revision History:

    tom.bonola@hp.com - 05/17/2006
    	- Converted original file to open source.

    tom.bonola@hp.com - 08/14/2003

        - Converted CpqCiAlignAddress() from macro to function.

        - Fixed pointer to integer conversion issue in the CpqCiCcbSetup()
          routine.  The .NET compiler did not like 64bit pointer type
          converted "down" to an integer (truncation).
--*/

#include "CpqCiTyp.h"
#include "CpqCi.h"
#include "CpqCiHlx.h"

/******************************************************************************
    Derived mask macros.
******************************************************************************/

#define _imask  Q->imask.uint              // index mask for head/tail
#define _merge  Q->merge.uint              // bits to merge on enqueue operation
#define _emask  (CPQCI_ENTRY_MASK >> 2)    // mask for entry less state bits
#define _omask  (CPQCI_ENTRY_MASK_O)       // mask for 'o'ccupied
#define _cmask  (CPQCI_ENTRY_MASK_C)       // mask for 'c'onsumed
#define _tmask  (_emask | _omask | _cmask) // mask for total entry bits

//**********************************************************
//* Simple conversion macros.
//**********************************************************

#define _bytes( _B )  (((u_t)(_B)) & 7)
#define _qwords( _B ) (_bytes(_B) ? ((((u_t)(_B)) >> 3) + 1) : (((u_t)(_B)) >> 3))



//**************************************************************
//* Critical Section Macros for queuing / dequeuing from Channel
//**************************************************************
void CpqCiLockAdapter(void);
void CpqCiUnlockAdapter(void);
void CpqCiLockChannel(int channel);
void CpqCiUnlockChannel(int channel);

#define ENTERCRITICALSECTION() CpqCiLockAdapter()
#define LEAVECRITICALSECTION() CpqCiUnlockAdapter()

CPQCIBUFFER   Temp1;
CPQCISIZE     Temp2;

//*********************************************
//*
//*     PUBLIC ROUTINES.
//*
//*********************************************


CPQCISIZE
CpqCiFifoSize(
    CPQCIENTRIES FifoEntries,
    CPQCIENTRIES *pAdjustedEntries 
    )

/*++

Description:

    This routine computes the minimum required size for the memory
    region that can be used as the fifo handle for a FIFO array of
    64-bit entries.

Parameters:

    FifoEntries - The requested number of entries for the fifo.
        This value is the guaranteed minimum requested entries
        the caller will have available for use.

    pAdjustedEntries - Location to place the adjusted number of
        entries that the handle will abstract.  The value
        returned in this location will be rounded up to a 2^n
        number of entries.  This parameter is ignored if the
        value for this location is NULL (0).

Return Value:

    The minimum size in bytes of the memory region required for
    a fifo handle of the requested number of fifo entries.

Caveats:

    The size returned accounts for any alignment that may need
    to occur by the CpqCiFifoSetup() routine.

--*/
{
    CPQCILENGTH adj = 0;
    CPQCILENGTH exp = 0;
    CPQCILENGTH len = 0;

    if( !FifoEntries )
        goto exitProc;

    /*******************************************************************
        The requested number of FIFO entries is the guaranteed
        minimum number of entries the caller will always have
        available.  We will add an additional entry to make the
        algorithms more efficient in handling FIFO boundary conditions.
    ********************************************************************/

    CpqCiLog2( FifoEntries + 1, adj, exp );

    len = (CPQCIFIFOHANDLESIZE + CPQCILINESIZE + (adj * 8));

exitProc:

    if( pAdjustedEntries ) 
        *pAdjustedEntries = adj;

    return len;
}

CPQCIFIFOBAR 
CpqCiFifoSetup(
    CPQCIBUFFER HandleMemory,
    CPQCIENTRIES FifoEntries,
    CPQCIBOOL Normal
    )

/*++

Description:

    This routine initializes the handle memory region 
    based on the requested fifo entries and valid data
    bits per fifo entry.

Parameters:

    HandleMemory - This is a caller supplied (and allocated) 
        memory region that will be initialized and used for
        the fifo handle.  The caller must not use this address
        for the queue handle because this routine may need to
        align it for optimal performance.

    FifoEntries - The requested number of entries for the fifo.
        This value is the guaranteed minimum requested entries
        the caller will have available for use.

    Normal - Dictates whether the CpqCiFifoEnqueue() routine
        immediately marks the entry as consumed when it gets placed
        into the fifo.  A value of !0 will cause the entry being
        enqueued to be immediately marked as consumed during the 
        enqueue operation.  Otherwise, the entry will not get marked
        consumed until another agent (i.e. HW) sets the state bit
        bit to consumed.  Keep in mind that a dequeue on items that
        have been enqueued without being marked consumed immediately
        results in an "empty" fifo condition.  A typical use for setting
        this parameter to a !0 value is when the Enqueuer and Dequeuer
        do not have another agent walking the fifo and marking entries
        as "Consumed".

Return Value:

    This routine returns the virtual address to the FIFOBAR for
    this fifo.  A value of NULL (0) indicates an error occurred during
    setup.

Caveats:

    The FifoEntries parameter should always be the value requested
    by the caller and not an "adjusted" value that may have been
    previously computed by an earlier call to CpqCiFifoSize().

--*/
{
    CPQCILENGTH adjustedentries;
    CPQCIFIFOBAR fifo = (void*)0;
    CPQCIFIFOHANDLE Q;
    u_t i;

    /*******************************************************************
        Compute the actual number of fifo entries.
    ********************************************************************/

    if( !CpqCiFifoSize(FifoEntries, &adjustedentries) )
        goto exitProc;

    /*******************************************************************
        Initialize the queue handle.  The dmask value represents the
        valid bits used to extract from the fifo entry.
    ********************************************************************/

    Q = CpqCiAlignAddress( HandleMemory, CPQCILINESIZE );

    Q->head = 0;
    Q->tail = 0;
    Q->reset.uint = 0;
    Q->rents.uint = FifoEntries;
    Q->imask.uint = adjustedentries - 1;
    Q->merge.uint = (Normal == CpqCiTrue) 
                        ? (CPQCI_ENTRY_MASK_O + CPQCI_ENTRY_MASK_C)
                        : (CPQCI_ENTRY_MASK_O)
                        ;

    for( i = 0; i < adjustedentries; i++ )
        Q->fifo[ i ] = 0;

    fifo = (void *)Q->fifo;

exitProc:
    return fifo;
}

CPQCIBOOL
CpqCiFifoEnqueue(
    CPQCIFIFOBAR Fifo,
    CPQCIENTRY Entry
    )

/*++

Description:

    This routine places the requested fifo entry onto the end of
    the requested fifo.

Parameters:

    Fifo - The base address of the target FIFO.

    Entry - The entry to place onto the target fifo.

Return Value:

    0  - Fifo FULL.
    !0 - Enqueue successful.

Caveats:

	None.
    
--*/
{
    CPQCIFIFOHANDLE Q = CpqCiFifoBar2FifoHandle(Fifo);
	CPQCIBOOL ok;

	//	We need exclusive execution.
	ENTERCRITICALSECTION();

	//	If the next entry is empty, then we can place the entry onto the tail
	//	of the FIFO queue.  Our definition of a full FIFO is a FIFO with only
	//	one empty entry.
	//
	//		Note: (Q->tail+1) points to the next fifo entry.
	//
	if(!(Q->fifo[(Q->tail+1) & _imask] & _omask))
	{
		Q->fifo[Q->tail & _imask] |= ((Entry & _emask) | _merge);
		Q->tail += 1;
		ok = CpqCiTrue;
	}
	else
		ok = CpqCiFalse;	// fifo full

	LEAVECRITICALSECTION();

	return ok;
}

CPQCIBOOL
CpqCiFifoDequeue(
    CPQCIFIFOBAR Fifo,
    CPQCIENTRY *pEntry
    )

/*++

Description:

    This routine removes the next available entry from the top of
    the requested fifo.

Parameters:

    Fifo - The base address of the target FIFO.

    pEntry - Location to place the entry removed from the fifo.  A value
        of NULL (0) will result in the removal of an entry with no assignment.

Return Value:

    0  - Fifo EMPTY.
    !0 - Dnqueue successful and is found in *pEntry if pEntry != NULL.

Caveats:

    When this routine returns TRUE (!0), the value in "pEntry" contains
    the application defined fifo entry bits and the 'O' & 'C' state bits
    set to 0.

--*/
{
    CPQCIFIFOHANDLE Q = CpqCiFifoBar2FifoHandle(Fifo);
	CPQCIBOOL ok;
	u64_t c;

	//	We need exclusive execution.
	ENTERCRITICALSECTION();

	c = Q->fifo[Q->head & _imask];

	if(c & _cmask)
	{
		if(pEntry)
			*pEntry = (CPQCIENTRY)(((u_t)(c)) & _emask);
		Q->fifo[Q->head & _imask] = ((c | _tmask) + 1);
		Q->head += 1;
		ok = CpqCiTrue;
	}
	else
		ok = CpqCiFalse;

	LEAVECRITICALSECTION();

	return ok;
}

CPQCIBOOL
CpqCiFifoEmpty(
    CPQCIFIFOBAR Fifo
    )

/*++

Description:

    This routine detects an empty FIFO condition and returns the 
    state of the FIFO to the caller.  This routine will attempt
    to keep the state of Fifo current as it progresses through
    the algorithm.

Parameters:

    Fifo - The base address of the target FIFO.

Return Value:

    0  - Fifo has items on it.
    !0 - Fifo is EMPTY.

Caveats:

    When this routine returns TRUE (!0), the value in "pEntry" contains
    the original fifo entry less the sequence number bits (set to 0's)
    and 'O' & 'C' state bits (set to 0's).

--*/
{
    CPQCIFIFOHANDLE Q = CpqCiFifoBar2FifoHandle(Fifo);
	CPQCIBOOL ok;

	//	We need exclusive execution.
	ENTERCRITICALSECTION();

	if(Q->fifo[Q->head & _imask] & _cmask)
		ok = CpqCiFalse;
	else
		ok = CpqCiTrue;
	
	LEAVECRITICALSECTION();

	return ok;
}

CPQCIBOOL
CpqCiFifoFull(
    CPQCIFIFOBAR Fifo
    )

/*++

Description:

    This routine detects a full FIFO condition and returns the 
    state of the FIFO to the caller.  This routine will attempt
    to keep the state of Fifo current as it progresses through
    the algorithm.

Parameters:

    Fifo - The base address of the target FIFO.

Return Value:

    0  - Fifo is not FULL.
    !0 - Fifo FULL

Caveats:

	None.

--*/
{
    CPQCIFIFOHANDLE Q = CpqCiFifoBar2FifoHandle(Fifo);
	CPQCIBOOL ok;

	//	We need exclusive execution.
	ENTERCRITICALSECTION();

	if(!(Q->fifo[(Q->tail+1)&_imask] & _omask))
    	ok = CpqCiFalse;  //fifo has available entries.
	else
        ok = CpqCiTrue;   //fifo full condition detected.

	LEAVECRITICALSECTION();

	return ok;
}

CPQCISTATUS
CpqCiCcbSetup(
    CPQCICCB     *Ccb,
    CPQCICHANNEL  ChannelNumber,
    CPQCIDOORBELL DoorbellBase,
    CPQCIFIFOBAR  SqFifoBar,
    CPQCIDESCBAR  SqDescBar,
    CPQCIENTRIES  SqDescs,
    CPQCISIZE     SqDescSize,
    CPQCICONTROL  SqGo,
    CPQCIFIFOBAR  RqFifoBar,
    CPQCIDESCBAR  RqDescBar,
    CPQCIENTRIES  RqDescs,
    CPQCISIZE     RqDescSize,
    CPQCICONTROL  RqGo
    )
{
    CPQCISTATUS error; // local error code
    CPQCICCB lccb;     // local ccb buffer
    u_t log2size;       // descriptor size as exponent
    u_t x;              // throw away

    /*******************************************************************
        Validate parameters.
    ********************************************************************/

    if( (SqDescs > CPQCI_MAX_DESCRIPTORS) || 
        (RqDescs > CPQCI_MAX_DESCRIPTORS) )
    {
        error = CpqCiErrBadDescriptorCount;
        goto exitProc;
    }

    if( (SqDescSize > CPQCI_MAX_DESCRIPTOR_SIZE) || 
        (RqDescSize > CPQCI_MAX_DESCRIPTOR_SIZE) )
    {
        error = CpqCiErrBadDescriptorSize;
        goto exitProc;
    }

    if( SqDescSize < CPQCI_MIN_DESCRIPTOR_SIZE )
        SqDescSize = CPQCI_MIN_DESCRIPTOR_SIZE;

    if( RqDescSize < CPQCI_MIN_DESCRIPTOR_SIZE )
        RqDescSize = CPQCI_MIN_DESCRIPTOR_SIZE;

    error = CpqCiSuccess;

    /*******************************************************************
        Clear out the local ccb buffer we will use to build the target
        CCB information.  We build a local copy first incase the caller
        passed us a CCB resident in un-cached memory (i.e. device ram).
    ********************************************************************/

    CpqCiFillBuffer( &lccb, 0, sizeof(CPQCICCB) );
    
    //  Update recv bars.
    lccb.RqFifo.ptr = RqFifoBar;
    lccb.RqDesc.ptr = RqDescBar;

    //  Compute descriptor log2 size and fifo index mask.
    if( RqDescs )
    {
        CpqCiLog2( RqDescSize, x, log2size );
        CpqCiFifoSize( RqDescs, &x );
        CpqCiCtrlSetAll( lccb.RqCtrl, log2size, x-1, RqDescs-1, 0, RqGo );
    }

    //  Update send bars.
    lccb.SqFifo.ptr = SqFifoBar;
    lccb.SqDesc.ptr = SqDescBar;

    //  Compute descriptor log2 size and fifo index mask.
    if( SqDescs )
    {
        CpqCiLog2( SqDescSize, x, log2size );
        CpqCiFifoSize( SqDescs, &x );
        CpqCiCtrlSetAll( lccb.SqCtrl, log2size, x-1, SqDescs-1, 0, SqGo );
    }

    lccb.DoorbellBase.ptr   = DoorbellBase;
    lccb.ChannelNumber.uint = ChannelNumber;

    //  Copy local ccb to target ccb.  Don't blow away
    //  target CCB's context field.
    {//tjb - 08/14/2003
        union {void *asptr; u64_t asint;} u;
        u.asint = 0; u.asptr = ((CPQCICCB *)(0))->Context;
        CpqCiCopyBuffer( Ccb, &lccb, u.asint );
    }

exitProc:
    return error;
}

CPQCIBOOL
CpqCiPacketEnqueue(
    CPQCICCB *Ccb,
    CPQCIFIFOTYPE TargetFifo,
    CPQCIPACKETID PacketId,
    CPQCILENGTH Length
    )

/*++

Description:

    This routine places the requested packet ID onto the
    end of the fifo represented by the Fifo handle.

Parameters:

    Ccb - The target channel control block.

    TargetFifo - Specifies the target FIFO within the target CCB.  Valid
        values are the following.

            CpqCiSendQ
            CpqCiRecvQ
    
    PacketID - The packet ID to be placed onto the tail of the fifo array.

    Length - The length in bytes of the packet specified by PacketId.

Return Value:

    0  - Fifo FULL.
    !0 - Enqueue successful.

Caveats:

    None.

--*/
{
    struct {
        CPQCITYPEQUAD Fifo;
        CPQCITYPEQUAD Desc;
        CPQCITYPEQUAD Ctrl;
    } *pfifo = (void *)&Ccb->SqFifo;

    CPQCIENTRY i = TargetFifo & CpqCiRecvQ;
    CPQCIENTRY e;

    CpqCiEntrySetAll( e, PacketId, _qwords(Length) );
    return CpqCiFifoEnqueue( pfifo[i].Fifo.ptr, e );
}

CPQCIBOOL
CpqCiPacketDequeue(
    CPQCICCB *Ccb,
    CPQCIFIFOTYPE TargetFifo,
    CPQCIPACKETID *pPacketId,
    CPQCILENGTH *pLength,
    CPQCIPACKET *pPacket
    )

/*++

Description:

    This routine removes the next available entry from the
    top of the fifo represented by the queue handle.

Parameters:

    Ccb - The target channel control block.

    TargetFifo - Specifies the target FIFO within the target CCB.  Valid
        values are the following.

            CpqCiSendQ
            CpqCiRecvQ
    
    pPacketId - Location to place the next packet descriptor#
        removed from the fifo.  The value placed in this location
        is only valid if this routine returns !0.

    pLength - Length in bytes represented by the descriptor specified
        in *pPacketId.

    pPacket - Pointer to the packet memory region relative to the
        caller's virtual address space.  This address corresponds
        to the *pPacketId value.

Return Value:

    0  - Fifo EMPTY.
    !0 - Dequeue successful and is found in *pPacketId and *pLength.

Caveats:

    None.

--*/
{
    struct {
        CPQCITYPEQUAD Fifo;
        CPQCITYPEQUAD Desc;
        CPQCITYPEQUAD Ctrl;
    } *pfifo = (void *)&Ccb->SqFifo;

    CPQCIENTRY i = TargetFifo & CpqCiRecvQ;
    CPQCIENTRY e;
    CPQCIPACKETID pid;

    if( CpqCiFifoDequeue( pfifo[i].Fifo.ptr, &e ) != CpqCiFalse )
    {
        pid = CpqCiEntryGetDescriptor( e );

        if( pPacketId ) *pPacketId = pid;
        if( pLength )   *pLength   = CpqCiEntryGetQwords( e ) << 3;
        if( pPacket )   *pPacket   = CpqCiMakePacketAddress(
                                        pfifo[i].Desc,
                                        pfifo[i].Ctrl,
                                        pid
                                        );
        return CpqCiTrue;
    }
    else
    {
        if( pLength )   *pLength = (CPQCILENGTH)0;
        if( pPacket )   *pPacket = (CPQCIPACKET)0;

        return CpqCiFalse;
    }
}

CPQCIBUFFER errormessages[] = 
{
    "Successful.",                                                              //CpqCiSuccess                 (0x00,  0)
    "CPQCIDRV driver is not loaded.",                                           //CpqCiErrNoDriver             (0x01,  1)
    "Unable to locate the device on this system.",                              //CpqCiErrNoDevice             (0x02,  2)
    "Memory allocation error.",                                                 //CpqCiErrNoMemory             (0x03,  3)
    "Bad name for the target connection object.",                               //CpqCiErrBadConnection        (0x04,  4)
    "The requested number of descriptors is not supported.",                    //CpqCiErrBadDescriptorCount   (0x05,  5)
    "The requested descriptor size is not supported.",                          //CpqCiErrBadDescriptorSize    (0x06,  6)
    "Unable to map device's base address register (BAR).",                      //CpqCiErrMapRegisters         (0x07,  7)
    "Unable to map device's doorbell base address.",                            //CpqCiErrMapDoorbells         (0x08,  8)
    "Unable to map device's shared ram base address.",                          //CpqCiErrMapDeviceram         (0x09,  9)
    "The device is not reporting a shared memory region.",                      //CpqCiErrNoDeviceMemory       (0x0A, 10)
    "The requested channel priority is not supported.",                         //CpqCiErrBadPriority          (0x0B, 11)
    "All channels are currently in use for the requested priority.",            //CpqCiErrNoMoreChannels       (0x0C, 12)
    "All connection objects are currently in use.",                             //CpqCiErrNoMoreConnections    (0x0D, 13)
    "Unable to allocate host memory for the target channel.",                   //CpqCiErrNoMoreChannelMemory  (0x0E, 14)
    "Unable to map channel's host memory into caller's address space.",         //CpqCiErrUnableToMapHostRam   (0x0F, 15)
    "Unable to map channel's send descriptor pool into caller's address space.",//CpqCiErrUnableToMapSendRam   (0x10, 16)
    "Unable to map channel's recv descriptor pool into caller's address space.",//CpqCiErrUnableToMapRecvRam   (0x11, 17)
    "Unable to map channel's doorbell aperature into caller's address space.",  //CpqCiErrUnableToMapDoorbell  (0x12, 18)
    "The driver is unable to process the specified object type.",               //CpqCiErrBadObjectType        (0x13, 19)
    "The send fifo is full.",                                                   //CpqCiErrFifoFull             (0x14, 20)
    "The receive fifo is empty.",                                               //CpqCiErrFifoEmpty            (0x15, 21)
    "The requested message size is larger than the target descriptor size.",    //CpqCiErrPartialTransfer      (0x16, 22)
    "The channel is still in use by 1 or more processes.",                      //CpqCiErrChannelStillInuse    (0x17, 23)
    "Access to the requested channel is denied.",                               //CpqCiErrAccessDenied         (0x18, 24)
    "The requested channel cannot be opened because it is marked for deletion.",//CpqCiErrChannelDeleted       (0x19, 25)
    "A valid key must be specified to open a channel.",                         //CpqCiErrNoKeySpecified       (0x1A, 26)
    "The device signature does not match this version of the CPQCIDRV driver.", //CpqCiErrInvalidAdapter       (0x1B, 27)
    "There are no channels pending service of the receive fifo.",               //CpqCiErrNoPendingChannels    (0x1C, 28)
    "A signal event could not be created for a channel.",                       //CpqCiErrUnableToCreateSignal (0x1D, 29)
    "A connection to the channel's event signal could not be obtained.",        //CpqCiErrUnableToObtainSignal (0x1E, 30)
    "The requested function is not implemented.",                               //CpqCiErrNotImplemented       (0x1F, 31)
    "The channel has no more pending waiters to alert.",                        //CpqCiErrNoMoreWaiters        (0x20, 32)
    "The requested function for the target channel only supports polling.",     //CpqCiErrUnableToWait         (0x21, 33)
    "The routine timedout without completing the request.",                     //CpqCiErrTimeout              (0x22, 34)
    "The channel interface is not initialized.",                                //CpqCiErrInterface            (0x23, 35)
    "The adapter is not responding to a channel request.",                      //CpqCiErrChannelNotResponding (0x24, 36)
    "The caller's environment context is not valid for this function.",         //CpqCiErrBadCallingContext    (0x25, 37)
    "The target channel has been disconnected due to an adapter reset.",        //CpqCiErrChannelDisconnected  (0x26, 38)
    "An invalid parameter was passed to this routine.",                         //CpqCiErrBadParameter         (0x27, 39)
    "Error code not supported."                                                 //CpqCiErrMax                  (0x28, 40)
};

CPQCIBUFFER
CpqCiStatusMessage(
    CPQCISTATUS StatusCode
    )

/*++

Description:

    This routine retrieves the message associated with the specified status code.

Parameters:

    StatusCode - CpqCi status code to convert into an ASCII text message.

Return Value:

    Pointer to the error related message.

Caveats:

    Status messages are statically allocated within this module and
    therefore must not be "free'd" by the caller.

--*/
{

    if( StatusCode >= CpqCiErrMax )
        StatusCode = CpqCiErrMax;

    return errormessages[StatusCode];
}
